/*
 *		      MALIGN.C  MS Johnson  24-06-1990
 *
 *
 *			     Copyright (c) 1990
 *
 *		Copyright (c) 1990 KRAM Enterprises, USA
 *		Copyright (c) 1990 KRAMIT Limited, UK
 *		Copyright (c) 1990 KRAMSHIT, Slovenia
 *		Copyright (c) 1990 KRAMSKU, Finlandia
 *
 *
 *	MAY BE FREELY DISTRIBUTED WITH THE PERMISSION OF THE AUTHOR
 *
 *	As with most things in life, use is at your own risk (except
 *	in the States where accidents and misfortunes are not permitted
 *	under the lawyers' Bill of Free Enterprise).
 *
 *	Mark S. Johnson, Modelling and Biocomputing Group, Center
 *	for Biotechnology, PO Box 123, FIN-20521 TURKU, FINLAND
 *	FAX: 358-21-633-8000
 *
 *	Multiple alignment program developed by the author at home
 *	and with his own resources. Thanks to Da-Fe Feng and Russ Doolittle
 *	for the original Fredman pairwise alignment code which has been
 *	only somewhat modified here.
 *
 *
 *	2.0 many bug fixes; msj
 *	2.1 fixed atoi for CONVEX version; msj
 */

                /*
		 *  file = Sequences to be aligned in PIR format.
		 */

#include <stdio.h>
#include <math.h>
#include "malign.h"
#include "mat11.h"

#define DEF_MAT   0			/* default matrix: birkbeck*/
#define DEF_PEN	 40			/* default non-negative gap penalty */

char code[MAX_SEQ][MAX_CDE];		/* seq ids */
char title[MAX_SEQ][MAX_TIT];		/* seq titles */
char seq[MAX_SEQ][MAX_LEN];		/* sequences */
int matrix[MAX_LEN][MAX_LEN];	/* score matrix to be traced */
int len[MAX_SEQ];			/* length of seqs; may include gaps! */
int ren[MAX_SEQ];			/* real length of seqs */
int nel_p[MAX_SEQ][MAX_SEQ];		/* aligned positions include overhgs*/
int nscore[MAX_SEQ][MAX_SEQ];	/* align score */
int s_flag;				/* segment smoothing if TRUE */
int seg;				/* segment length to smooth over */
int temp[2][MAX_LEN];		/* temp storage for two seqs */
int family[MAX_SEQ][MAX_SEQ];	/* treed members */
float fscore[MAX_SEQ][MAX_SEQ];		/* norm align score */
float rscore[MAX_SEQ][MAX_SEQ];		/* raw align score */
float fpid[MAX_SEQ][MAX_SEQ];		/* seq ids */
float fdif[MAX_SEQ][MAX_SEQ];		/* seq diffs */

int p_f;		/* internal flag to control average over template*/
int m_f;		/* internal flag to average or not over template*/
int r_f;		/* full alignment on; trace out if 1; 0 = off */

int mf_flag=0;		/* places signal "^" about final align */

int read_um();
void comp_um();
int pair_um();
float norm_um();
void print_um();
void dist_um();
void tree_um();
float pidf_um();
float pids_um();
float pdif_um();
int score_um();


main(ac,av)
int ac;
char *av[];
{

FILE *file, *fopen();

register int i, j;

int ns;			/* number of sequences */
int pen;		/* pen value to use */
int mat;		/* mat number to use */
int score;		/* alignemnt score */
int t_len;		/* length of alignment */
int list1[MAX_SEQ];	/* list of seqs to align as set */
int list2[MAX_SEQ];	/* list of seqs to align as set */
int nlist1,nlist2;	/* number in set to be aligned with another set */
int lenj,lenk;		/* temp length of two sets of seqs */
int p_ct=0;		/* controls the working dots printout */
int set;		/* counts input args */
float tscore;		/* sum of normalized scores */
char Temp[20];		/* used to read args */


/*
 *	open sequences file for reading
 */

	
if((file=fopen(av[1],"r"))==NULL)
	{
	fprintf(stderr,"\n\n%s: Cannot open %s! (FATAL)",av[0],av[1]);
	fprintf(stderr,"\nUsage: %s file_of_seqs [ -seg=# ] [ -M# ] ",av[0]);
	fprintf(stderr,"[ -P# ] [ -O ] [ > output ]");
	fprintf(stderr,"\n\nFile_of_seqs -- sequences in PIR format!");
	fprintf(stderr,"\n\tMaximum number of sequences: %d",MAX_SEQ);
	fprintf(stderr,"\n\tMaximum length of sequences: %d\n",MAX_LEN-52);
	fprintf(stderr,"\n-seg=#  : smooth over segment # residues long; try 6");
	fprintf(stderr,"\n-M# : scoring matrix");
	fprintf(stderr,"; -P# : gap penalty");
	fprintf(stderr,"; -O  : for binary alignment output\n");
	fprintf(stderr,"\nMatrix vals:    %s= 0  pen = 40; (DEFAULT)\n\n",m_names[0]);
	for(i=1;i<MAX_MAT;i++)
		{
		if(i%2!=0)fprintf(stderr,"%-20.20s	= %d",m_names[i],i);
		else fprintf(stderr,"		%-20.20s	= %d\n",m_names[i],i);
		}
	fprintf(stderr,"\n%s\n",VER); 
	fprintf(stderr,"%s\n",MAT_VER); 
	exit(1);
	}
else 	{
	printf("\n***************************************");
	printf("************************************");
	printf("\n*************************************");
	printf("**************************************");
	printf("\n* %s: %s ",av[0],VER);
	printf("\n**************************************");
	printf("*************************************");
	printf("\n**************************************");
	printf("*************************************");
	}

/*
 *	enter matrix number; penalty and r_f (=1 for all pairwise aligns).
 */

mat=DEF_MAT;
pen=DEF_PEN;
s_flag=FALSE;

r_f=0;

set=2;

while(set<ac)
	{
	if(strncmp(av[set],"-seg=",5)==0)
		{
		i=5;
		while(av[set][i]!='\0')
			{
			Temp[i-5]=av[set][i];
			i++;
			}
		Temp[i-5]='\0';
		seg=atoi(Temp);
		s_flag=TRUE;
		set++;
		}
	else if(strncmp(av[set],"-M",2)==0)
		{
		i=2;
		while(av[set][i]!='\0')
			{
			Temp[i-2]=av[set][i];
			i++;
			}
		Temp[i-2]='\0';
		mat=atoi(Temp);
		set++;
		}
	else if(strncmp(av[set],"-P",2)==0)
		{
		i=2;
		while(av[set][i]!='\0')
			{
			Temp[i-2]=av[set][i];
			i++;
			}
		Temp[i-2]='\0';
		pen=atoi(Temp);
		set++;
		}
	else if(strncmp(av[set],"-O",2)==0)
		{
		r_f=1;
		set++;
		}
	else	{
		printf("\nUnrecogized option: %s \n",av[set]);
		exit(49);
		}
	}

if(s_flag==TRUE)printf("\n\nSmoothing over segments %d long",seg);
else printf("\n");
printf("\nMatrix:  %s",m_names[mat]);
printf("    Penalty: %d",pen);

printf("\n\n**************************************");
printf("*************************************\n\n");

if(mat>=MAX_MAT)
	{
	fprintf(stderr,"\n\nMatrix value %d exceeds listings (FATAL)\n\n",mat);
	fprintf(stderr,"\n\nUsage: %s file_of_seqs [ -seg=# ] [ -M# ] ",av[0]);
	fprintf(stderr,"[ -P# ] [ -O ] [ > output ]");
	fprintf(stderr,"\n\nFile_of_seqs -- sequences in PIR format!");
	fprintf(stderr,"\n\tMaximum number of sequences: %d",MAX_SEQ);
	fprintf(stderr,"\n\tMaximum length of sequences: %d\n",MAX_LEN-52);
	fprintf(stderr,"\n-seg=#  : smooth over segment # residues long; try 6");
	fprintf(stderr,"\n-M# : scoring matrix");
	fprintf(stderr,"\n-P# : gap penalty");
	fprintf(stderr,"\n-O  : for binary alignment output\n");
	fprintf(stderr,"\nMatrix vals:    %s= 0  pen = 40; (DEFAULT)\n\n",m_names[0]);
	for(i=1;i<MAX_MAT;i++)
		{
		if(i%2!=0)fprintf(stderr,"%-20.20s	= %d",m_names[i],i);
		else fprintf(stderr,"		%-20.20s	= %d\n",m_names[i],i);
		}
	fprintf(stderr,"\n%s\n",VER); 
	fprintf(stderr,"%s\n",MAT_VER); 
	exit(3);
	}

/*
 *	read the sequences into seq[0 ... n]
 */

ns=read_um(file);

/*
 *	only a pair so turn on printout
 */

if(ns==2)
	{
	r_f=1;
	mf_flag=1;	/* print ^ to signify final alignment */
	}

/*
 *	make pairwise alignments
 */

p_f=0;
m_f=0;

if(r_f==0)printf("\n\nWorking\n ");

for(i=0;i<ns-1;i++)
	{
	for(j=i+1;j<ns;j++)
		{

		/*
		 * comp_um produces matrix for pair_um to align from.
		 * m_f = 0 here; do not average over more than one
		 *            sequence.
		 */

		comp_um(i,j,mat,&lenj,&lenk);

		/*
 		 * pair_um aligns from the values given in matrix
		 * r_f=1 here; means trace out alignment
		 * and return score
		 */

		score=pair_um(lenj,lenk,pen,&t_len,1);

		/*
		 *	score separately; score from pair_um may be off
		 *	by a multiple of the gap penalty if the sequences
		 *	had gaps introduced into the sequences prior to
		 *	their comparison by malign. 1 will indicate
		 *	data from pair_um is located in temp[2][len]
		 *	rather than seq[][].
		 */

		score=score_um(i,j,t_len,mat,pen,1);

		nscore[i][j]=score;
		nscore[j][i]=score;

		/*
		 * norm_um normalizes the score by dividing by
		 * the average of the two self-self comparisons
		 * returns fscore
		 */


		fscore[i][j]=norm_um(i,j,t_len,score,mat,1);
		fscore[j][i]=fscore[i][j];

		if(r_f==1)printf("\n_______________________________");
		if(r_f==1)
		printf("____________________________________________\n\n");
		if(r_f==1)printf("Similarity score= %6d ",score);
		if(r_f==1)printf("Normalized score= %8.1f\n",fscore[i][j]);
		if(r_f==0)
			{
			if(p_ct>0 && (p_ct+1)%38 ==0)printf("\n");
			printf(". ");
			fflush(stdout);
			p_ct++;
			}

		/*
		 * print_um prints out in PIR format each set of aligned
		 * sequences in list1 and list2; obviously only if r_f==1
		 */

		list1[0]=i;nlist1=1;
		list2[0]=j;nlist2=1;
		if(r_f==1)print_um(list1,nlist1,list2,nlist2,t_len);


		/*
		 *	percentage identities -- pairwise only
		 */
		
		fpid[i][j]=pids_um(i,j,t_len);
		fpid[j][i]=fpid[i][j];

		}
	}

if(r_f==0)printf("\nDone\n");

fflush(stdout);

/*
 *	print the fids
 */

printf("\n\n******************************************************");
printf("*********************");
printf("\n************************ Pairwise comparison statistics:");
printf(" ******************");
printf("\n********************************************************");
printf("*******************");

printf("\n\nPercentage identities:\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[i]);
	for(j=0;j<ns;j++)
		{
		if(i==j)fpid[i][j]=100.0;
		if(j>0 && j%10==0)printf("\n            ");
		printf("%6.1f",fpid[i][j]);
		}
	}

printf("\n");

printf("\n\nNormalized alignment scores (NAS = 100.0 * Sreal/Smax):\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[i]);
	for(j=0;j<ns;j++)
		{
		if(i==j)fscore[i][j]=100.0;
		if(j>0 && j%10==0)printf("\n            ");
		printf("%6.1f",fscore[i][j]);
		}
	}

printf("\n");

/*
 *	exit at this point if only a pair of seqs.
 */

if(ns==2)
	{
	printf("\n\n");
	exit(0);
	}

/*
 *	create distances from norm align score -- pairwise only or scores
 */

dist_um(ns);

/*
 *	sort by hierarchy tree and multiply align as tree is constructed
 */

p_f=0;
r_f=1;
tree_um(ns,mat,pen,&t_len);

/*
 *	calculate percentage id for all pairs; then flush
 */

for(i=0;i<ns-1;i++)
	{
	for(j=i+1;j<ns;j++)
		{
		fpid[i][j]=pidf_um(i,j,t_len);
		fpid[j][i]=fpid[i][j];
		}
	}


/*
 *	calculate percentage difference for all pairs; then flush
 */

for(i=0;i<ns-1;i++)
	{
	for(j=i+1;j<ns;j++)
		{
		fdif[i][j]=pdif_um(i,j,t_len);
		fdif[j][i]=fdif[i][j];
		}
	}

/*
 *	calculate norm align score for all pairs; then flush
 */

tscore=0.0;
for(i=0;i<ns;i++)
	{
	for(j=i;j<ns;j++)
		{
		score=score_um(i,j,t_len,mat,pen,0);
		rscore[i][j]= (float)score;
		rscore[j][i]= (float)score;
		if(i!=j)tscore+= rscore[i][j];
		fscore[i][j]=norm_um(i,j,t_len,score,mat,0);
		fscore[j][i]=fscore[i][j];
		}
	}

printf("\n\n******************************************************");
printf("**********************");
printf("\n************************ Multiple comparison statistics:");
printf(" *******************");
printf("\n********************************************************");
printf("********************");

printf("\n\nPercentage identities:\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
		if(i==j)fpid[family[0][i]-1][family[0][j]-1]=100.0;
		if(j>0 && j%10==0)printf("\n            ");
		printf("%6.1f",fpid[family[0][i]-1][family[0][j]-1]);
		}
	}
printf("\n");


printf("\n\nDifferences:\n\n");
/*printf("\n\n100 * normalized differences:\n\n");*/
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
		if(i==j)fdif[family[0][i]-1][family[0][j]-1]=0.0;
		if(j>0 && j%10==0)printf("\n            ");
		printf("%6.0f",fdif[family[0][i]-1][family[0][j]-1]);
		}
	}
printf("\n");


printf("\n\nLengths of alignment including overhangs, gaps:\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
		if(j>0 && j%10==0)printf("\n            ");
		if(i==j) printf("%6d",ren[family[0][i]-1]);
		else printf("%6d",nel_p[family[0][i]-1][family[0][j]-1]);
		}
	}
printf("\n");


printf("\n\nDistances from identities (D = -100.0 ln pids/100):\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
		if(i==j)fpid[family[0][i]-1][family[0][j]-1]=100.0;
		if(j>0 && j%10==0)printf("\n            ");
		printf("%6.1f",fabs(-100.0 * log((fpid[family[0][i]-1][family[0][j]-1]/100.0))));
		}
	}
printf("\n");


printf("\n\nRaw alignment scores / 100 \n");
printf("[Overall (sum of off-diagonal elements): %8.1f]:\n\n",tscore/100.0);
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
	/*	if(i==j)fscore[family[0][i]-1][family[0][j]-1]=100.0;*/
		if(j>0 && j%10==0)printf("\n            ");
		printf(" %5.1f",rscore[family[0][i]-1][family[0][j]-1]/100.0);
		}
	}
printf("\n");


printf("\n\nNormalized alignment scores (NAS = 100.0 * Sreal/Smax):\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
	/*	if(i==j)fscore[family[0][i]-1][family[0][j]-1]=100.0;*/
		if(j>0 && j%10==0)printf("\n            ");
		printf(" %5.1f",fscore[family[0][i]-1][family[0][j]-1]);
		}
	}
printf("\n");


printf("\n\nDistances based on NAS scores [-100.0 * ln(NAS/100.0)]:\n\n");
printf("            ");
for(i=0;i<ns;i++)
	{
	if(i>0 && i%10==0)printf("\n            ");
	printf(" [%3d]",i+1);
	}
printf("\n");
for(i=0;i<ns;i++)
	{
	printf("\n%3d-%-8.8s",i+1,code[family[0][i]-1]);
	for(j=0;j<ns;j++)
		{
	/*	if(i==j)fscore[family[0][i]-1][family[0][j]-1]=100.0;*/
		if(j>0 && j%10==0)printf("\n            ");
		printf(" %5.1f",
		fabs(-100.0 * log(fscore[family[0][i]-1][family[0][j]-1]/100.0)));
		}
	}
printf("\n");

/*
 *	exit
 */

printf("\n\n********************************************************");
printf("********************");
printf("\n\n");

fclose(file);

exit(0);

}
